"""
rohmu

Copyright (c) 2016 Ohmu Ltd
Copyright (c) 2022 Aiven, Helsinki, Finland. https://aiven.io/
See LICENSE for details
"""
from .common.models import StorageModel
from .errors import InvalidConfigurationError
from .notifier.interface import Notifier
from .object_storage.base import BaseTransfer
from typing import Any, Dict, Mapping, Type

IO_BLOCK_SIZE = 2**20  # 1 MiB
STORAGE_TYPE = "storage_type"
NOTIFIER_TYPE = "notifier_type"
Config = Mapping[str, Any]


def get_class_for_transfer(obj_store: Config) -> Type[BaseTransfer]:
    storage_type = obj_store[STORAGE_TYPE]
    if storage_type == "azure":
        from .object_storage.azure import AzureTransfer

        return AzureTransfer
    elif storage_type == "google":
        from .object_storage.google import GoogleTransfer

        return GoogleTransfer
    elif storage_type == "sftp":
        from .object_storage.sftp import SFTPTransfer

        return SFTPTransfer
    elif storage_type == "local":
        from .object_storage.local import LocalTransfer

        return LocalTransfer
    elif storage_type == "s3":
        from .object_storage.s3 import S3Transfer

        return S3Transfer
    elif storage_type == "swift":
        from .object_storage.swift import SwiftTransfer

        return SwiftTransfer

    raise InvalidConfigurationError("unsupported storage type {0!r}".format(storage_type))


def get_class_for_notifier(notifier_config: dict) -> Type[Notifier]:
    notifier_type = notifier_config[NOTIFIER_TYPE]
    if notifier_type == "http":
        from .notifier.http import BackgroundHTTPNotifier

        return BackgroundHTTPNotifier
    raise InvalidConfigurationError("unsupported storage type {0!r}".format(notifier_type))


def get_notifier(notifier_config: dict) -> Notifier:
    notificer_class = get_class_for_notifier(notifier_config)
    notifier_config = notifier_config.copy()
    notifier_config.pop(NOTIFIER_TYPE)
    return notificer_class(**notifier_config)


def get_transfer_model(storage_config: Config) -> Type[StorageModel]:
    storage_class = get_class_for_transfer(storage_config)
    storage_config = dict(storage_config)
    storage_config.pop(STORAGE_TYPE)
    notifier_config = storage_config.pop("notifier", None)
    notifier = None
    if notifier_config is not None:
        notifier = get_notifier(notifier_config)

    model = storage_class.config_model(**storage_config, notifier=notifier)
    return model


def get_transfer(storage_config: Config) -> BaseTransfer:
    storage_class = get_class_for_transfer(storage_config)
    model = get_transfer_model(storage_config)
    return storage_class.from_model(model)



"""
rohmu - google cloud object store interface

Copyright (c) 2016 Ohmu Ltd
Copyright (c) 2022 Aiven, Helsinki, Finland. https://aiven.io/
See LICENSE for details
"""
# pylint: disable=import-error, no-name-in-module

from ..common.models import ProxyInfo, StorageModel
from ..dates import parse_timestamp
from ..errors import FileNotFoundFromStorageError, InvalidConfigurationError
from ..notifier.interface import Notifier
from .base import (
    BaseTransfer,
    get_total_memory,
    IncrementalProgressCallbackType,
    IterKeyItem,
    KEY_TYPE_OBJECT,
    KEY_TYPE_PREFIX,
    ProgressProportionCallbackType,
)
from contextlib import contextmanager
from googleapiclient.discovery import build
from googleapiclient.errors import HttpError
from googleapiclient.http import build_http, MediaFileUpload, MediaIoBaseDownload, MediaIoBaseUpload, MediaUpload
from http.client import IncompleteRead
from io import BytesIO, FileIO
from oauth2client import GOOGLE_TOKEN_URI
from oauth2client.client import GoogleCredentials
from typing import Optional

import codecs
import errno

# NOTE: this import is not needed per-se, but it's imported here first to point the
# user to the most important possible missing dependency
import googleapiclient  # noqa pylint: disable=unused-import
import httplib2
import json
import logging
import os
import random
import socket
import ssl
import time

try:
    from oauth2client.service_account import ServiceAccountCredentials

    ServiceAccountCredentials_from_dict = ServiceAccountCredentials.from_json_keyfile_dict
except ImportError:
    from oauth2client.service_account import _ServiceAccountCredentials

    def ServiceAccountCredentials_from_dict(credentials, scopes=None):
        if scopes is None:
            scopes = []
        return _ServiceAccountCredentials(
            service_account_id=credentials["client_id"],
            service_account_email=credentials["client_email"],
            private_key_id=credentials["private_key_id"],
            private_key_pkcs8_text=credentials["private_key"],
            scopes=scopes,
        )


# Silence Google API client verbose spamming
logging.getLogger("googleapiclient.discovery_cache").setLevel(logging.ERROR)
logging.getLogger("googleapiclient").setLevel(logging.WARNING)
logging.getLogger("oauth2client").setLevel(logging.WARNING)

# googleapiclient download performs some 3-4 times better with 50 MB chunk size than 5 MB chunk size;
# but decrypting/decompressing big chunks needs a lot of memory so use smaller chunks on systems with less
# than 2 GB RAM
DOWNLOAD_CHUNK_SIZE = 1024 * 1024 * 5 if get_total_memory() < 2048 else 1024 * 1024 * 50
UPLOAD_CHUNK_SIZE = 1024 * 1024 * 5


def get_credentials(credential_file=None, credentials=None):
    if credential_file:
        return GoogleCredentials.from_stream(credential_file)

    if credentials and credentials["type"] == "service_account":
        return ServiceAccountCredentials_from_dict(
            credentials,
            scopes=["https://www.googleapis.com/auth/cloud-platform"],
        )

    if credentials and credentials["type"] == "authorized_user":
        return GoogleCredentials(
            access_token=None,
            client_id=credentials["client_id"],
            client_secret=credentials["client_secret"],
            refresh_token=credentials["refresh_token"],
            token_expiry=None,
            token_uri=GOOGLE_TOKEN_URI,
            user_agent="pghoard",
        )

    return GoogleCredentials.get_application_default()


def base64_to_hex(b64val):
    if isinstance(b64val, str):
        b64val = b64val.encode("ascii")
    rawval = codecs.decode(b64val, "base64")
    hexval = codecs.encode(rawval, "hex")
    return hexval.decode("ascii")


class Config(StorageModel):
    project_id: str
    bucket_name: str
    credential_file: Optional[str] = None
    credentials: Optional[dict] = None
    proxy_info: Optional[ProxyInfo] = None
    prefix: Optional[str] = None


class GoogleTransfer(BaseTransfer[Config]):
    config_model = Config

    def __init__(
        self,
        project_id,
        bucket_name,
        credential_file=None,
        credentials=None,
        prefix=None,
        proxy_info=None,
        notifier: Optional[Notifier] = None,
    ) -> None:
        super().__init__(prefix=prefix, notifier=notifier)
        self.project_id = project_id
        self.proxy_info = proxy_info
        self.google_creds = get_credentials(credential_file=credential_file, credentials=credentials)
        self.gs = self._init_google_client()
        self.gs_object_client = None
        self.bucket_name = self.get_or_create_bucket(bucket_name)
        self.log.debug("GoogleTransfer initialized")

    def _init_google_client(self):
        start_time = time.monotonic()
        delay = 2
        while True:
            http = build_http()
            if self.proxy_info:
                if self.proxy_info.get("type") == "socks5":
                    proxy_type = httplib2.socks.PROXY_TYPE_SOCKS5
                else:
                    proxy_type = httplib2.socks.PROXY_TYPE_HTTP

                http.proxy_info = httplib2.ProxyInfo(
                    proxy_type,
                    self.proxy_info["host"],
                    self.proxy_info["port"],
                    proxy_user=self.proxy_info.get("user"),
                    proxy_pass=self.proxy_info.get("pass"),
                )

            http = self.google_creds.authorize(http)

            try:
                # sometimes fails: httplib2.ServerNotFoundError: Unable to find the server at www.googleapis.com
                # https://googleapis.github.io/google-api-python-client/docs/dyn/storage_v1.html
                return build("storage", "v1", http=http)
            except (httplib2.ServerNotFoundError, socket.timeout):
                if time.monotonic() - start_time > 600:
                    raise

            # retry on DNS issues
            time.sleep(delay)
            delay = delay * 2

    @contextmanager
    def _object_client(self, *, not_found=None):
        """(Re-)initialize object client if required, handle 404 errors gracefully and reset the client on
        server errors.  Server errors have been shown to be caused by invalid state in the client and do not
        seem to be resolved without resetting."""
        if self.gs_object_client is None:
            if self.gs is None:
                self.gs = self._init_google_client()
            # https://googleapis.github.io/google-api-python-client/docs/dyn/storage_v1.objects.html
            self.gs_object_client = self.gs.objects()  # pylint: disable=no-member

        try:
            yield self.gs_object_client
        except HttpError as ex:
            if ex.resp["status"] == "404" and not_found is not None:
                raise FileNotFoundFromStorageError(not_found)
            if ex.resp["status"] >= "500" and ex.resp["status"] <= "599":
                self.log.error("Received server error %r, resetting Google API client", ex.resp["status"])
                self.gs = None
                self.gs_object_client = None
            raise

    def _retry_on_reset(self, request, action):
        retries = 60
        retry_wait = 2
        while True:
            try:
                return action()
            except (IncompleteRead, HttpError, ssl.SSLEOFError, socket.timeout, OSError, socket.gaierror) as ex:
                # Note that socket.timeout and ssl.SSLEOFError inherit from OSError
                # and the order of handling the errors here needs to be correct
                if not retries:
                    raise
                elif isinstance(ex, (IncompleteRead, socket.timeout, ssl.SSLEOFError, BrokenPipeError)):
                    pass  # just retry with the same sleep amount
                elif isinstance(ex, HttpError):
                    # https://cloud.google.com/storage/docs/json_api/v1/status-codes
                    # https://cloud.google.com/storage/docs/exponential-backoff
                    if ex.resp["status"] not in ("429", "500", "502", "503", "504"):  # pylint: disable=no-member
                        raise
                    retry_wait = min(10.0, max(1.0, retry_wait * 2) + random.random())
                # httplib2 commonly fails with Bad File Descriptor and Connection Reset
                elif isinstance(ex, OSError) and ex.errno not in [errno.EAGAIN, errno.EBADF, errno.ECONNRESET]:
                    raise
                # getaddrinfo sometimes fails with "Name or service not known"
                elif isinstance(ex, socket.gaierror) and ex.errno != socket.EAI_NONAME:
                    raise

                self.log.warning("%s failed: %s (%s), retrying in %.2fs", action, ex.__class__.__name__, ex, retry_wait)

            # we want to reset the http connection state in case of error
            if request and hasattr(request, "http"):
                request.http.connections.clear()  # reset connection cache

            retries -= 1
            time.sleep(retry_wait)

    def copy_file(self, *, source_key, destination_key, metadata=None, **_kwargs):
        source_object = self.format_key_for_backend(source_key)
        destination_object = self.format_key_for_backend(destination_key)
        body = {}
        if metadata is not None:
            body["metadata"] = metadata

        with self._object_client(not_found=source_key) as clob:
            # https://googleapis.github.io/google-api-python-client/docs/dyn/storage_v1.objects.html#copy
            request = clob.copy(
                body=body,
                destinationBucket=self.bucket_name,
                destinationObject=destination_object,
                sourceBucket=self.bucket_name,
                sourceObject=source_object,
            )
            result = self._retry_on_reset(request, request.execute)
            if result.get("size", None) is not None:
                self.notifier.object_copied(key=destination_key, size=int(result["size"]), metadata=metadata)

    def get_metadata_for_key(self, key):
        path = self.format_key_for_backend(key)
        with self._object_client(not_found=path) as clob:
            return self._metadata_for_key(clob, path)

    def _metadata_for_key(self, clob, key):
        # https://googleapis.github.io/google-api-python-client/docs/dyn/storage_v1.objects.html#get
        req = clob.get(bucket=self.bucket_name, object=key)
        obj = self._retry_on_reset(req, req.execute)
        return obj.get("metadata", {})

    def _unpaginate(self, domain, initial_op, *, on_properties):
        """Iterate thru the request pages until all items have been processed"""
        request = initial_op(domain)
        while request is not None:
            result = self._retry_on_reset(request, request.execute)
            for on_property in on_properties:
                items = result.get(on_property)
                if items is not None:
                    yield on_property, items
            request = domain.list_next(request, result)

    def iter_key(
        self, key, *, with_metadata=True, deep=False, include_key=False  # pylint: disable=unused-argument, unused-variable
    ):
        path = self.format_key_for_backend(key, trailing_slash=not include_key)
        self.log.debug("Listing path %r", path)
        with self._object_client() as clob:

            def initial_op(domain):
                if deep:
                    kwargs = {}
                else:
                    kwargs = {"delimiter": "/"}
                # https://googleapis.github.io/google-api-python-client/docs/dyn/storage_v1.objects.html#list
                return domain.list(bucket=self.bucket_name, prefix=path, **kwargs)

            for property_name, items in self._unpaginate(clob, initial_op, on_properties=["items", "prefixes"]):
                if property_name == "items":
                    for item in items:
                        if item["name"].endswith("/"):
                            self.log.warning("list_iter: directory entry %r", item)
                            continue  # skip directory level objects

                        yield IterKeyItem(
                            type=KEY_TYPE_OBJECT,
                            value={
                                "name": self.format_key_from_backend(item["name"]),
                                "size": int(item["size"]),
                                "last_modified": parse_timestamp(item["updated"]),
                                "metadata": item.get("metadata", {}),
                                "md5": base64_to_hex(item["md5Hash"]),
                            },
                        )
                elif property_name == "prefixes":
                    for prefix in items:
                        yield IterKeyItem(type=KEY_TYPE_PREFIX, value=self.format_key_from_backend(prefix).rstrip("/"))
                else:
                    raise NotImplementedError(property_name)

    def delete_key(self, key):
        path = self.format_key_for_backend(key)
        self.log.debug("Deleting key: %r", path)
        with self._object_client(not_found=path) as clob:
            # https://googleapis.github.io/google-api-python-client/docs/dyn/storage_v1.objects.html#delete
            req = clob.delete(bucket=self.bucket_name, object=path)
            self._retry_on_reset(req, req.execute)
            self.notifier.object_deleted(key)

    def get_contents_to_file(self, key, filepath_to_store_to, *, progress_callback: ProgressProportionCallbackType = None):
        fileobj = FileIO(filepath_to_store_to, mode="wb")
        done = False
        metadata = {}
        try:
            metadata = self.get_contents_to_fileobj(key, fileobj, progress_callback=progress_callback)
            done = True
        finally:
            fileobj.close()
            if not done:
                os.unlink(filepath_to_store_to)
        return metadata

    def get_contents_to_fileobj(self, key, fileobj_to_store_to, *, progress_callback: ProgressProportionCallbackType = None):
        path = self.format_key_for_backend(key)
        self.log.debug("Starting to fetch the contents of: %r to %r", path, fileobj_to_store_to)
        next_prog_report = 0.0
        last_log_output = 0.0
        with self._object_client(not_found=path) as clob:
            # https://googleapis.github.io/google-api-python-client/docs/dyn/storage_v1.objects.html#get_media
            req = clob.get_media(bucket=self.bucket_name, object=path)
            download = MediaIoBaseDownload(fileobj_to_store_to, req, chunksize=DOWNLOAD_CHUNK_SIZE)
            done = False
            while not done:
                status, done = self._retry_on_reset(getattr(download, "_request", None), download.next_chunk)
                if status:
                    progress_pct = status.progress() * 100
                    now = time.monotonic()
                    if (now - last_log_output) >= 5.0:
                        self.log.debug("Download of %r: %d%%", path, progress_pct)
                        last_log_output = now

                    if progress_callback and progress_pct > next_prog_report:
                        progress_callback(progress_pct, 100)
                        next_prog_report = progress_pct + 0.1
            if progress_callback:
                progress_callback(100, 100)
            return self._metadata_for_key(clob, path)

    def get_contents_to_string(self, key):
        path = self.format_key_for_backend(key)
        self.log.debug("Starting to fetch the contents of: %r", path)
        with self._object_client(not_found=path) as clob:
            # https://googleapis.github.io/google-api-python-client/docs/dyn/storage_v1.objects.html#get_media
            req = clob.get_media(bucket=self.bucket_name, object=path)
            data = self._retry_on_reset(req, req.execute)
            return data, self._metadata_for_key(clob, path)

    def get_file_size(self, key):
        path = self.format_key_for_backend(key)
        with self._object_client(not_found=path) as clob:
            # https://googleapis.github.io/google-api-python-client/docs/dyn/storage_v1.objects.html#get
            req = clob.get(bucket=self.bucket_name, object=path)
            obj = self._retry_on_reset(req, req.execute)
            return int(obj["size"])

    def _upload(
        self, upload, key, metadata, extra_props, cache_control, upload_progress_fn: IncrementalProgressCallbackType = None
    ):
        path = self.format_key_for_backend(key)
        self.log.debug("Starting to upload %r", path)
        body = {"metadata": metadata}
        if extra_props:
            body.update(extra_props)
        if cache_control is not None:
            body["cacheControl"] = cache_control

        last_log_output = 0.0
        with self._object_client() as clob:
            # https://googleapis.github.io/google-api-python-client/docs/dyn/storage_v1.objects.html#insert
            req = clob.insert(bucket=self.bucket_name, name=path, media_body=upload, body=body)
            response = None
            while response is None:
                status, response = self._retry_on_reset(req, req.next_chunk)
                if status:
                    now = time.monotonic()
                    if (now - last_log_output) >= 5.0:
                        self.log.debug(
                            "Upload of %r to %r: %d%%, %s bytes",
                            upload,
                            path,
                            status.progress() * 100,
                            status.resumable_progress,
                        )
                        last_log_output = now

                    if upload_progress_fn:
                        upload_progress_fn(status.resumable_progress)
        return response

    # pylint: disable=arguments-differ
    def store_file_from_memory(self, key, memstring, metadata=None, extra_props=None, cache_control=None, mimetype=None):
        data = BytesIO(memstring)
        upload = MediaIoBaseUpload(data, mimetype or "application/octet-stream", chunksize=UPLOAD_CHUNK_SIZE, resumable=True)
        sanitized_metadata = self.sanitize_metadata(metadata)
        result = self._upload(upload, key, sanitized_metadata, extra_props, cache_control=cache_control)
        self.notifier.object_created(key=key, size=int(result.get("size", len(memstring))), metadata=sanitized_metadata)
        return result

    # pylint: disable=arguments-differ
    def store_file_from_disk(
        self,
        key,
        filepath,
        metadata=None,  # pylint: disable=arguments-differ, unused-variable
        multipart=None,
        cache_control=None,
        mimetype=None,
        progress_fn: ProgressProportionCallbackType = None,
        *,
        extra_props=None,  # pylint: disable=arguments-differ, unused-variable
    ):
        size = os.path.getsize(filepath)
        mimetype = mimetype or "application/octet-stream"
        upload = MediaFileUpload(filepath, mimetype, chunksize=UPLOAD_CHUNK_SIZE, resumable=True)
        sanitized_metadata = self.sanitize_metadata(metadata)
        result = self._upload(
            upload,
            key,
            sanitized_metadata,
            extra_props,
            cache_control=cache_control,
            upload_progress_fn=self._incremental_to_proportional_progress(cb=progress_fn, size=size),
        )
        self.notifier.object_created(key=key, size=size, metadata=sanitized_metadata)
        if progress_fn:
            progress_fn(size, size)
        return result

    def store_file_object(
        self,
        key,
        fd,
        *,
        cache_control=None,
        metadata=None,
        mimetype=None,
        upload_progress_fn: IncrementalProgressCallbackType = None,
    ):
        mimetype = mimetype or "application/octet-stream"
        sanitized_metadata = self.sanitize_metadata(metadata)
        result = self._upload(
            MediaStreamUpload(fd, chunk_size=UPLOAD_CHUNK_SIZE, mime_type=mimetype, name=key),
            key,
            sanitized_metadata,
            None,
            cache_control=cache_control,
            upload_progress_fn=upload_progress_fn,
        )
        self.notifier.object_created(key=key, size=int(result["size"]), metadata=sanitized_metadata)
        return result

    def get_or_create_bucket(self, bucket_name):
        """Look up the bucket if it already exists and try to create the
        bucket in case it doesn't.  Note that we can't just always try to
        unconditionally create the bucket as Google imposes a strict rate
        limit on bucket creation operations, even if it doesn't result in a
        new bucket.

        Quietly handle the case where the bucket already exists to avoid
        race conditions.  Note that we'll get a 400 Bad Request response for
        invalid bucket names ("Invalid bucket name") as well as for invalid
        project ("Invalid argument"), try to handle both gracefully."""
        start_time = time.time()
        gs_buckets = self.gs.buckets()  # pylint: disable=no-member
        try:
            request = gs_buckets.get(bucket=bucket_name)
            self._retry_on_reset(request, request.execute)
            self.log.debug("Bucket: %r already exists, took: %.3fs", bucket_name, time.time() - start_time)
        except HttpError as ex:
            if ex.resp["status"] == "404":
                pass  # we need to create it
            elif ex.resp["status"] == "403":
                raise InvalidConfigurationError("Bucket {0!r} exists but isn't accessible".format(bucket_name))
            else:
                raise
        else:
            return bucket_name

        try:
            req = gs_buckets.insert(project=self.project_id, body={"name": bucket_name})
            self._retry_on_reset(req, req.execute)
            self.log.debug("Created bucket: %r successfully, took: %.3fs", bucket_name, time.time() - start_time)
        except HttpError as ex:
            error = json.loads(ex.content.decode("utf-8"))["error"]
            if error["message"].startswith("You already own this bucket"):
                self.log.debug("Bucket: %r already exists, took: %.3fs", bucket_name, time.time() - start_time)
            elif error["message"] == "Invalid argument.":
                raise InvalidConfigurationError("Invalid project id {0!r}".format(self.project_id))
            elif error["message"].startswith("Invalid bucket name"):
                raise InvalidConfigurationError("Invalid bucket name {0!r}".format(bucket_name))
            else:
                raise

        return bucket_name


class MediaStreamUpload(MediaUpload):
    """Support streaming arbitrary amount of data from non-seekable object supporting read method."""

    def __init__(self, fd, *, chunk_size, mime_type, name):
        self._data = b""
        self._next_chunk = b""
        self._chunk_size = chunk_size
        self._fd = fd
        self._mime_type = mime_type
        self._name = name
        self._position = None

    def chunksize(self):
        return self._chunk_size

    def mimetype(self):
        return self._mime_type

    def size(self):
        self.peek()
        if len(self._next_chunk) < self.peeksize:
            # The total file size should be returned if we have hit the final chunk.
            return (self._position or 0) + len(self._data) + len(self._next_chunk)
        return None

    def resumable(self):
        return True

    @property
    def peeksize(self):
        # Using 1 extra byte to avoid perfectly aligned file
        return self._chunk_size + 1

    def peek(self):
        """try to top up some data into _next_chunk"""
        if len(self._next_chunk) < self.peeksize:
            # top-up the _next_chunk
            self._next_chunk = self._read_bytes(self.peeksize - len(self._next_chunk), initial_data=self._next_chunk)

    # second parameter is length but baseclass incorrectly names it end
    def getbytes(self, begin, length):  # pylint: disable=arguments-differ
        if begin < (self._position or 0):
            msg = "Requested position {} for {!r} precedes already fulfilled position {}".format(
                begin, self._name, self._position
            )
            raise IndexError(msg)
        elif begin > (self._position or 0) + len(self._data):
            msg = "Requested position {} for {!r} has gap from previous position {} and {} byte chunk".format(
                begin, self._name, self._position, len(self._data)
            )
            raise IndexError(msg)

        if self._position is None or begin == self._position + len(self._data):
            if length <= len(self._next_chunk):
                self._data = self._next_chunk[:length]
                self._next_chunk = self._next_chunk[length:]
            else:
                self._data = self._read_bytes(length - len(self._next_chunk), initial_data=self._next_chunk)
                self._next_chunk = b""
        elif begin != self._position or length > len(self._data):
            retain_chunk = self._data[begin - self._position :]
            bytes_remaining = length - len(retain_chunk)
            if 0 < bytes_remaining <= len(self._next_chunk):
                self._data = retain_chunk + self._next_chunk[:bytes_remaining]
                self._next_chunk = self._next_chunk[bytes_remaining:]
            elif bytes_remaining > len(self._next_chunk):
                retain_chunk += self._next_chunk
                self._next_chunk = b""
                self._data = self._read_bytes(length - len(retain_chunk), initial_data=retain_chunk)

        self.peek()
        self._position = begin
        return self._data

    def has_stream(self):
        return False

    def stream(self):
        raise NotImplementedError

    def _read_bytes(self, length, *, initial_data=None):
        bytes_remaining = length
        read_results = []
        if initial_data:
            read_results.append(initial_data)
        while bytes_remaining > 0:
            data = self._fd.read(bytes_remaining)
            if data:
                read_results.append(data)
                bytes_remaining -= len(data)
            else:
                break

        if not read_results:
            return b""
        elif len(read_results) == 1:
            return read_results[0]
        else:
            return b"".join(read_results)



"""
rohmu - content encryption

Copyright (c) 2016 Ohmu Ltd
See LICENSE for details
"""

from . import IO_BLOCK_SIZE
from .filewrap import FileWrap, Sink, Stream
from cryptography.hazmat.backends import default_backend
from cryptography.hazmat.primitives import serialization
from cryptography.hazmat.primitives.asymmetric import padding
from cryptography.hazmat.primitives.ciphers import algorithms, Cipher, modes
from cryptography.hazmat.primitives.hashes import SHA1, SHA256
from cryptography.hazmat.primitives.hmac import HMAC

import cryptography
import cryptography.hazmat.backends.openssl.backend
import io
import logging
import os
import struct

if cryptography.__version__ < "1.6":  # type: ignore
    # workaround for deadlock https://github.com/pyca/cryptography/issues/2911
    cryptography.hazmat.backends.openssl.backend.activate_builtin_random()

FILEMAGIC = b"pghoa1"
AES_BLOCK_SIZE = 16


class EncryptorError(Exception):
    """EncryptorError"""


class Encryptor:
    def __init__(self, rsa_public_key_pem):
        if not isinstance(rsa_public_key_pem, bytes):
            rsa_public_key_pem = rsa_public_key_pem.encode("ascii")
        self.rsa_public_key = serialization.load_pem_public_key(rsa_public_key_pem, backend=default_backend())
        self.cipher = None
        self.authenticator = None

    def update(self, data):
        ret = b""
        if self.cipher is None:
            key = os.urandom(16)
            nonce = os.urandom(16)
            auth_key = os.urandom(32)
            self.cipher = Cipher(algorithms.AES(key), modes.CTR(nonce), backend=default_backend()).encryptor()
            self.authenticator = HMAC(auth_key, SHA256(), backend=default_backend())
            pad = padding.OAEP(mgf=padding.MGF1(algorithm=SHA1()), algorithm=SHA1(), label=None)
            cipherkey = self.rsa_public_key.encrypt(key + nonce + auth_key, pad)
            ret = FILEMAGIC + struct.pack(">H", len(cipherkey)) + cipherkey
        cur = self.cipher.update(data)
        self.authenticator.update(cur)
        if ret:
            return ret + cur
        else:
            return cur

    def finalize(self):
        if self.cipher is None:
            return b""  # empty plaintext input yields empty encrypted output

        ret = self.cipher.finalize()
        self.authenticator.update(ret)
        ret += self.authenticator.finalize()
        self.cipher = None
        self.authenticator = None
        return ret


class EncryptorFile(FileWrap):
    def __init__(self, next_fp, rsa_public_key_pem):
        super().__init__(next_fp)
        self.key = rsa_public_key_pem
        self.encryptor = Encryptor(self.key)
        self.offset = 0
        self.state = "OPEN"

    def flush(self):
        self._check_not_closed()
        self.next_fp.flush()

    def close(self):
        if self.state == "CLOSED":
            return
        final = self.encryptor.finalize()
        self.encryptor = None
        self.next_fp.write(final)
        super().close()

    def writable(self):
        """True if this stream supports writing"""
        self._check_not_closed()
        return True

    def write(self, data):
        """Encrypt and write the given bytes"""
        self._check_not_closed()
        if not data:
            return 0
        enc_data = self.encryptor.update(data)
        self.next_fp.write(enc_data)
        self.offset += len(data)
        return len(data)


class EncryptorStream(Stream):
    """Non-seekable stream of data that adds encryption on top of given source stream"""

    def __init__(self, src_fp, rsa_public_key_pem):
        super().__init__(src_fp)
        self._encryptor = Encryptor(rsa_public_key_pem)

    def _process_chunk(self, data):
        return self._encryptor.update(data)

    def _finalize(self):
        return self._encryptor.finalize()


class Decryptor:
    def __init__(self, rsa_private_key_pem):
        if not isinstance(rsa_private_key_pem, bytes):
            rsa_private_key_pem = rsa_private_key_pem.encode("ascii")
        self.rsa_private_key = serialization.load_pem_private_key(
            data=rsa_private_key_pem, password=None, backend=default_backend()
        )
        self.cipher = None
        self.authenticator = None
        self._cipher_key_len = None
        self._header_size = None
        self._footer_size = 32

    def expected_header_bytes(self):
        if self._header_size is not None:
            return 0
        return self._cipher_key_len or 8

    def header_size(self):
        return self._header_size

    def footer_size(self):
        return self._footer_size

    def process_header(self, data):
        if self._cipher_key_len is None:
            if data[0:6] != FILEMAGIC:
                raise EncryptorError("Invalid magic bytes")
            self._cipher_key_len = struct.unpack(">H", data[6:8])[0]
        else:
            pad = padding.OAEP(mgf=padding.MGF1(algorithm=SHA1()), algorithm=SHA1(), label=None)
            try:
                plainkey = self.rsa_private_key.decrypt(data, pad)
            except AssertionError:
                raise EncryptorError("Decrypting key data failed")
            if len(plainkey) != 64:
                raise EncryptorError("Integrity check failed")
            key = plainkey[0:16]
            nonce = plainkey[16:32]
            auth_key = plainkey[32:64]
            self._header_size = 8 + len(data)

            self.cipher = Cipher(algorithms.AES(key), modes.CTR(nonce), backend=default_backend()).decryptor()
            self.authenticator = HMAC(auth_key, SHA256(), backend=default_backend())

    def process_data(self, data):
        if not data:
            return b""
        self.authenticator.update(data)
        return self.cipher.update(data)

    def finalize(self, footer):
        if footer != self.authenticator.finalize():
            raise EncryptorError("Integrity check failed")
        result = self.cipher.finalize()
        self.cipher = None
        self.authenticator = None
        return result


class DecryptorFile(FileWrap):
    def __init__(self, next_fp, rsa_private_key_pem):
        super().__init__(next_fp)
        self._key = rsa_private_key_pem
        self.log = logging.getLogger(self.__class__.__name__)
        self._decryptor = None
        self._crypted_size = None
        self._boundary_block = None
        self._plaintext_size = None
        # Our actual plain-text read offset. seek may change self.offset to something
        # else temporarily but we keep _decrypt_offset intact until we actually do a
        # read in case the caller just called seek in order to then immediately seek back
        self._decrypt_offset = None
        self.offset = None
        self._reset()

    def _reset(self):
        self._decryptor = Decryptor(self._key)
        self._crypted_size = self._file_size(self.next_fp)
        self._boundary_block = None
        self._plaintext_size = None
        self._decrypt_offset = 0
        # Plaintext offset
        self.offset = 0
        self.state = "OPEN"

    @classmethod
    def _file_size(cls, file):
        current_offset = file.seek(0, os.SEEK_SET)
        file_end_offset = file.seek(0, os.SEEK_END)
        file.seek(current_offset, os.SEEK_SET)
        return file_end_offset

    def _initialize_decryptor(self):
        if self._plaintext_size is not None:
            return
        while True:
            required_bytes = self._decryptor.expected_header_bytes()
            if not required_bytes:
                break
            self._decryptor.process_header(self._read_raw_exactly(required_bytes))
        self._plaintext_size = self._crypted_size - self._decryptor.header_size() - self._decryptor.footer_size()

    def _read_raw_exactly(self, required_bytes):
        data = self.next_fp.read(required_bytes)
        while data and len(data) < required_bytes:
            next_chunk = self.next_fp.read(required_bytes - len(data))
            if not next_chunk:
                break
            data += next_chunk
        if not data or len(data) != required_bytes:
            raise EncryptorError("Failed to read {} bytes of header or footer data".format(required_bytes))
        return data

    def _move_decrypt_offset_to_plaintext_offset(self):
        if self._decrypt_offset == self.offset:
            return
        seek_to = self.offset
        if self._decrypt_offset > self.offset:
            self.log.warning("Negative seek from %d to %d, must re-initialize decryptor", self._decrypt_offset, self.offset)
            self._reset()
            self._initialize_decryptor()
        discard_bytes = seek_to - self._decrypt_offset
        self.offset = self._decrypt_offset
        while discard_bytes > 0:
            data = self._read_block(discard_bytes)
            discard_bytes -= len(data)

    def _read_all(self):
        full_data = bytearray()
        while True:
            data = self._read_block(IO_BLOCK_SIZE)
            if not data:
                return bytes(full_data)
            full_data.extend(data)

    def _read_block(self, size):
        if self._crypted_size == 0:
            return b""

        self._initialize_decryptor()

        if self.offset == self._plaintext_size:
            return b""

        self._move_decrypt_offset_to_plaintext_offset()

        # If we have an existing boundary block, fulfil the read entirely from that
        if self._boundary_block:
            size = min(size, len(self._boundary_block) - self.offset % AES_BLOCK_SIZE)
            data = self._boundary_block[self.offset % AES_BLOCK_SIZE : self.offset % AES_BLOCK_SIZE + size]
            if self.offset % AES_BLOCK_SIZE + size == len(self._boundary_block):
                self._boundary_block = None
            data_len = len(data)
            self.offset += data_len
            self._decrypt_offset += data_len
            return data

        # Only serve multiples of AES_BLOCK_SIZE whenever possible to keep things simpler
        read_size = size
        if self.offset + max(AES_BLOCK_SIZE, size) >= self._plaintext_size:
            read_size = self._plaintext_size - self.offset
        elif size > AES_BLOCK_SIZE and size % AES_BLOCK_SIZE != 0 and self.offset + size < self._plaintext_size:
            read_size = size - size % AES_BLOCK_SIZE
        elif size < AES_BLOCK_SIZE:
            read_size = AES_BLOCK_SIZE

        encrypted = self._read_raw_exactly(read_size)
        decrypted = self._decryptor.process_data(encrypted)
        if self.offset + read_size == self._plaintext_size:
            footer = self._read_raw_exactly(self._decryptor.footer_size())
            last_part = self._decryptor.finalize(footer)
            if last_part:
                decrypted += last_part

        if size < AES_BLOCK_SIZE:
            self._boundary_block = decrypted
            return self._read_block(size)
        decrypted_len = len(decrypted)
        self.offset += decrypted_len
        self._decrypt_offset += decrypted_len
        return decrypted

    def close(self):
        super().close()
        self._decryptor = None

    def read(self, size=-1):
        """Read up to size decrypted bytes"""
        self._check_not_closed()
        if self.state == "EOF" or size == 0:
            return b""
        elif size < 0:
            return self._read_all()
        else:
            return self._read_block(size)

    def readable(self):
        """True if this stream supports reading"""
        self._check_not_closed()
        return self.state in ["OPEN", "EOF"]

    def seek(self, offset, whence=0):
        self._check_not_closed()
        self._initialize_decryptor()
        if whence == os.SEEK_SET:
            if offset != self.offset:
                if offset > self._plaintext_size:
                    raise io.UnsupportedOperation("DecryptorFile does not support seeking beyond EOF")
                if offset < 0:
                    raise ValueError("negative seek position")
                self.offset = offset
            return self.offset
        elif whence == os.SEEK_CUR:
            if offset != 0:
                raise io.UnsupportedOperation("can't do nonzero cur-relative seeks")
            return self.offset
        elif whence == os.SEEK_END:
            if offset != 0:
                raise io.UnsupportedOperation("can't do nonzero end-relative seeks")
            self.offset = self._plaintext_size
            return self.offset
        else:
            raise ValueError("Invalid whence value")

    def seekable(self):
        """True if this stream supports random access"""
        self._check_not_closed()
        return True


class DecryptSink(Sink):
    def __init__(self, next_sink, file_size, encryption_key_data):
        super().__init__(next_sink)
        if file_size <= 0:
            raise ValueError("Invalid file_size: " + str(file_size))
        self.data_bytes_received = 0
        self.data_size = file_size
        self.decryptor = Decryptor(encryption_key_data)
        self.file_size = file_size
        self.footer = b""
        self.header = b""

    def _extract_encryption_footer_bytes(self, data):
        expected_data_bytes = self.data_size - self.data_bytes_received
        if len(data) > expected_data_bytes:
            self.footer += data[expected_data_bytes:]
            data = data[:expected_data_bytes]
        return data

    def _process_encryption_header(self, data):
        if not data or not self.decryptor.expected_header_bytes():
            return data
        if self.header:
            data = self.header + data
            self.header = None
        offset = 0
        while self.decryptor.expected_header_bytes() > 0:
            header_bytes = self.decryptor.expected_header_bytes()
            if header_bytes + offset > len(data):
                self.header = data[offset:]
                return b""
            self.decryptor.process_header(data[offset : offset + header_bytes])
            offset += header_bytes
        data = data[offset:]
        self.data_size = self.file_size - self.decryptor.header_size() - self.decryptor.footer_size()
        return data

    def write(self, data):
        written = len(data)
        data = self._process_encryption_header(data)
        if not data:
            return written
        data = self._extract_encryption_footer_bytes(data)
        self.data_bytes_received += len(data)
        if data:
            data = self.decryptor.process_data(data)
        if len(self.footer) == self.decryptor.footer_size():
            final_data = self.decryptor.finalize(self.footer)
            if final_data:
                data += final_data
        if not data:
            return written
        self._write_to_next_sink(data)
        return written



# Copyright (c) 2023 Aiven, Helsinki, Finland. https://aiven.io/


from rohmu.notifier.interface import Notifier
from typing import Optional

import enum
import pydantic


class ProxyType(str, enum.Enum):
    socks5 = "socks5"
    http = "http"


class RohmuModel(pydantic.BaseModel):
    class Config:
        # As we're keen to both export and decode json, just using
        # enum values for encode/decode is much saner than the default
        # enumname.value (it is also slightly less safe but oh well)
        use_enum_values = True

        # Extra values should be errors, as they are most likely typos
        # which lead to grief when not detected. However, if we ever
        # start deprecating some old fields and not wanting to parse
        # them, this might need to be revisited.
        extra = "forbid"

        # Validate field default values too
        validate_all = True

        # Validate also assignments
        # validate_assignment = True
        # TBD: Figure out why this doesn't work in some unit tests;
        # possibly the tests themselves are broken


class ProxyInfo(RohmuModel):
    host: str
    port: int
    type: ProxyType
    user: Optional[str]
    password: Optional[str] = pydantic.Field(None, alias="pass")


class StorageModel(pydantic.BaseModel):
    notifier: Optional[Notifier] = None

    class Config:
        use_enum_values = True
        # Notifier Type does not have validation (yet)
        arbitrary_types_allowed = True
        # extra fields would not be accepted by corresponding __init__ methods anyways
        extra_forbid = True



from pathlib import Path
from rohmu.atomic_opener import atomic_opener

import errno
import os
import pytest
import time


def _verify_file_not_created_and_dir_not_polluted(output_file: Path):
    assert os.listdir(output_file.parent) == []
    assert not output_file.exists()


def test_error_thrown_if_final_path_parent_doesnt_exist(tmp_path: Path):
    with pytest.raises(IOError):
        with atomic_opener(tmp_path / "nonexistingdir" / "final_path", mode="w"):
            pass


def test_error_mode_doesnt_contain_write(tmp_path: Path):
    with pytest.raises(ValueError):
        with atomic_opener(tmp_path, mode="r"):  # type: ignore
            pass


def test_file_is_atomically_created_only_after_function_execution_is_over(tmp_path: Path):
    data_block = "x" * 100_000_000
    # manually tested with block_count of 1000 but it seems overkill to do it every time and it can hog
    # the testing infra quite a bit.
    # block_count = 1000
    block_count = 1
    size = len(data_block)

    output_file = tmp_path / "something"
    try:
        _verify_file_not_created_and_dir_not_polluted(output_file)
        with atomic_opener(output_file, mode="w") as f:
            inode_inside = os.stat(f.fileno()).st_ino
            _verify_file_not_created_and_dir_not_polluted(output_file)
            for unused_counter in range(block_count):
                f.write(data_block)
                f.flush()
                _verify_file_not_created_and_dir_not_polluted(output_file)
            timestamp_before_exiting = time.monotonic()

        # this must be super-fast since atomic move is taking place. Makes more sense with
        # a large - 100ish - block count
        assert (time.monotonic() - timestamp_before_exiting) < 0.1
        assert output_file.exists()
        assert len(os.listdir(output_file.parent)) == 1
        assert output_file.stat().st_size == size * block_count
        assert output_file.stat().st_ino == inode_inside
    finally:
        # the file can be big, we want to delete it ASAP
        try:
            output_file.unlink()
        except FileNotFoundError:
            pass


def test_file_is_never_created_if_function_breaks(tmp_path: Path):
    output_file = tmp_path / "something"

    try:
        _verify_file_not_created_and_dir_not_polluted(output_file)
        with atomic_opener(output_file, mode="w") as f:
            _verify_file_not_created_and_dir_not_polluted(output_file)
            f.write("aaaaaaaaaa")
            f.flush()
            _verify_file_not_created_and_dir_not_polluted(output_file)
            time.sleep(1.0)
            _verify_file_not_created_and_dir_not_polluted(output_file)
            raise ValueError("crash")
        pytest.fail("codepath should never be taken")
    except ValueError:
        _verify_file_not_created_and_dir_not_polluted(output_file)


def test_file_is_fully_written_if_visible(tmp_path: Path):
    output_file = tmp_path / "something"

    def linkhook():
        assert output_file.exists()
        assert output_file.read_text() == "aaaaaaaaaa"

    try:
        _verify_file_not_created_and_dir_not_polluted(output_file)
        with atomic_opener(output_file, mode="w", _after_link_hook=linkhook) as f:
            _verify_file_not_created_and_dir_not_polluted(output_file)
            f.write("aaaaaaaaaa")
            _verify_file_not_created_and_dir_not_polluted(output_file)
            time.sleep(1.0)
            _verify_file_not_created_and_dir_not_polluted(output_file)

    except ValueError:
        _verify_file_not_created_and_dir_not_polluted(output_file)


def test_open_for_writing_text_opens_proper_encoding_file(tmp_path: Path):
    final_path = tmp_path / "file"
    with atomic_opener(final_path, encoding="iso-8859-1", mode="w") as f:
        f.write("à")
    assert final_path.read_bytes() == b"\xe0"


def test_open_for_writing_bytes_properly_writes_bytes(tmp_path: Path):
    final_path = tmp_path / "file"
    with atomic_opener(final_path, mode="wb") as f:
        f.write(b"\xe0")
    assert final_path.read_text("iso-8859-1") == "à"


def test_no_fd_leak_if_fdopen_fails_because_of_wrong_encoding(tmp_path: Path):
    final_path = tmp_path / "file"
    opened_fd: list[int] = []
    try:
        with atomic_opener(final_path, mode="w", encoding="unknownencoding", _fd_spy=opened_fd.append):
            pass
        pytest.fail("should fail, encoding is wrong")
    except LookupError:
        try:
            os.fstat(opened_fd[0])
            pytest.fail("should fail, file descriptor must be invalid")
        except OSError as e:
            if e.errno != errno.EBADF:
                raise
            # descriptor is invalid, all ok


def test_no_fd_leak_if_fdopen_fails_because_of_unknown_mode(tmp_path: Path):
    final_path = tmp_path / "file"
    opened_fd: list[int] = []
    try:
        with atomic_opener(final_path, mode="somethingrandomw", encoding="ascii", _fd_spy=opened_fd.append):
            pass
        pytest.fail("should fail, mode is wrong")
    except ValueError:
        try:
            os.fstat(opened_fd[0])
            pytest.fail("should fail, file descriptor must be invalid")
        except OSError as e:
            if e.errno != errno.EBADF:
                raise
            # descriptor is invalid, all ok



"""
Copyright (c) 2020 Aiven Ltd
See LICENSE for details

Test that StatsClient works as advertised.

This could be broader, but at least minimal functionality is tested here.

"""
from __future__ import annotations

from rohmu.common import statsd
from typing import Any

import asyncio
import pytest
import socket


class _Protocol(asyncio.DatagramProtocol):
    def __init__(self, queue: asyncio.Queue):
        self.received_queue = queue

    def datagram_received(self, data: bytes, addr: tuple[str | Any, int]) -> None:
        self.received_queue.put_nowait(data)


@pytest.mark.asyncio
async def test_statsd():
    loop = asyncio.get_running_loop()
    received = asyncio.Queue()

    sock = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
    sock.bind(("", 0))

    transport, _ = await loop.create_datagram_endpoint(lambda: _Protocol(received), sock=sock)
    sock = transport.get_extra_info("socket")
    port = sock.getsockname()[-1]

    # Ensure that no config = no action
    c = statsd.StatsClient(config=None)
    c.increase("foo")

    c = statsd.StatsClient(config=statsd.StatsdConfig(port=port))
    c.increase("bar")

    data = await received.get()
    assert data == b"bar:1|c"

    with c.timing_manager("sync_timing"):
        pass
    data = await received.get()
    assert data.startswith(b"sync_timing,success=1:") and data.endswith(b"|ms")

    async with c.async_timing_manager("async_timing"):
        pass
    data = await received.get()
    assert data.startswith(b"async_timing,success=1:") and data.endswith(b"|ms")



"""Copyright (c) 2022 Aiven, Helsinki, Finland. https://aiven.io/"""
from datetime import datetime
from io import BytesIO
from tempfile import NamedTemporaryFile
from types import ModuleType
from unittest.mock import MagicMock, patch

import pytest
import sys


@pytest.fixture(scope="module", name="swift_module")
def fixture_swift_module() -> ModuleType:
    with patch.dict(sys.modules, {"swiftclient": MagicMock()}):
        import rohmu.object_storage.swift

    return rohmu.object_storage.swift


def test_store_file_from_disk(swift_module: ModuleType) -> None:
    notifier = MagicMock()
    connection = MagicMock()
    swift_module.client.Connection.return_value = connection
    transfer = swift_module.SwiftTransfer(
        user="testuser",
        key="testkey",
        container_name="test_container",
        auth_url="http://auth.example.com",
        notifier=notifier,
    )
    test_data = b"test-data"
    metadata = {"Content-Length": len(test_data), "some-date": datetime(2022, 11, 15, 18, 30, 58, 486644)}
    with NamedTemporaryFile() as tmpfile:
        tmpfile.write(test_data)
        tmpfile.flush()
        transfer.store_file_from_disk(key="test_key1", filepath=tmpfile.name, metadata=metadata)

    connection.put_object.assert_called()
    notifier.object_created.assert_called_once_with(
        key="test_key1", size=len(test_data), metadata={"Content-Length": "9", "some-date": "2022-11-15 18:30:58.486644"}
    )


def test_store_file_object(swift_module: ModuleType) -> None:
    notifier = MagicMock()
    connection = MagicMock()
    swift_module.client.Connection.return_value = connection
    transfer = swift_module.SwiftTransfer(
        user="testuser",
        key="testkey",
        container_name="test_container",
        auth_url="http://auth.example.com",
        notifier=notifier,
    )
    test_data = b"test-data"
    file_object = BytesIO(test_data)
    metadata = {"Content-Length": len(test_data), "some-date": datetime(2022, 11, 15, 18, 30, 58, 486644)}
    transfer.store_file_object(key="test_key2", fd=file_object, metadata=metadata)

    connection.put_object.assert_called()
    notifier.object_created.assert_called_once_with(
        key="test_key2", size=len(test_data), metadata={"Content-Length": "9", "some-date": "2022-11-15 18:30:58.486644"}
    )


def test_iter_key_with_empty_key(swift_module: ModuleType) -> None:
    notifier = MagicMock()
    connection = MagicMock(get_container=MagicMock(return_value=[None, {}]))
    swift_module.client.Connection.return_value = connection
    transfer = swift_module.SwiftTransfer(
        user="testuser",
        key="testkey",
        container_name="test_container",
        auth_url="http://auth.example.com",
        notifier=notifier,
    )
    list(transfer.iter_key(""))
    transfer.conn.get_container.assert_called_with("test_container", prefix="", full_listing=True, delimiter="/")



#!/usr/bin/env python3

# This file enables editable installs with `pip install --editable .`
# https://setuptools.readthedocs.io/en/latest/setuptools.html#setup-cfg-only-projects

import setuptools

setuptools.setup()
